"""Common types used in pystv."""
from dataclasses import dataclass

from icontract import invariant


class PyStvError(Exception):
    """Error in PyStv."""


@dataclass
class RaceMetadata:
    """Parameters defining a single race."""

    race_name: str
    num_winners: int
    names: list[str]

    def __str__(self):
        cands = ",".join(self.names)
        return (
            f"race: {self.race_name}\n"
            f"num_winners: {self.num_winners}\n"
            f"candidates: {cands}"
        )


@invariant(lambda self: len(self.metadata.names) >= max(map(max, self.ballots)))
@invariant(lambda self: len(self.ballots) == len(self.votes))
@dataclass
class RaceData:
    """Voting data for a single race.

    ballots: A list of candidate rankings.  Each list[int] is an ordering of a subset
             candidate indexes, with the index refering to the list of candidates in
             `metadata`.
    votes: A list of the same length as `ballots` which denotes the number of votes
           corresponding to each candidate ranking.  The sum of `votes` corresponds
           to the total number of votes cast.

    The two main ways to use this class:
    - Each ballot corresponds to one person's vote.  Multiple ballots can be identical,
      and all the entries in votes are 1.  This usage is the most verbose and uses more
      memory, but does correspond to the typical understanding of a ballot.
    - Each ballot corresponds to a unique ordering of the candidates.  All ballots
      are unique.  The entry in `votes` corresponding to a given ballot indicates the
      number of people who voted in that ordering.  This usage is the most compact and
      uses less memory.

    For example, the following would be identical in a 2 candidate race with 7 voters:
    - ballots: [[1,2], [2], [2,1], [2,1], [2], [1], [2, 1]]
      votes: [1, 1, 1, 1, 1, 1, 1]
    - ballots: [1], [2], [1, 2], [2, 1]
      votes: [1, 2, 1, 3]
    """

    metadata: RaceMetadata
    ballots: list[list[int]]
    votes: list[int]


@dataclass
class RoundResult:
    """The full results of a single-transferable voting round.

    count: The votes for each candidate.
    elected: List of candidate ids which have won the election by this round.
    eliminated: List of candidate indices which have lost the election by this round.
    transfers: Counts of votes that have been transferred during this round.  It is a
               nested dict of src_cand_index -> tgt_cand_index -> vote_count

    """

    count: list[float]
    elected: list[int]
    eliminated: list[int]
    transfers: dict[int, dict[int, float]]


@invariant(lambda self: len(set(len(r.count) for r in self.rounds)) < 2)
@invariant(
    lambda self: not self.rounds
    or (len(self.rounds[0].count) == len(self.metadata.names) + 1)
)
@dataclass
class RaceResult:
    """The list of all round results for a single race."""

    metadata: RaceMetadata
    rounds: list[RoundResult]

    def __str__(self):
        def rnd2str(round_result):
            details = []
            names_all = ["<exhausted>"] + self.metadata.names
            for idx, (name, cnt) in enumerate(zip(names_all, round_result.count)):
                name_detail = f"{name}: {cnt}"
                if idx in round_result.elected:
                    name_detail += " +"
                if idx in round_result.eliminated:
                    name_detail += " -"
                details.append(name_detail)
                # Add transfers
            return "\n".join(details)

        round_info = []
        for rnd, round_results in enumerate(self.rounds):
            round_info.extend([f"Round {rnd}:", rnd2str(round_results)])
        rounds = "\n".join(round_info)
        return f"{self.metadata}\n{rounds}"
